<?php

namespace Drupal\search_api_solr_shards;

use Drupal\Component\Utility\Unicode;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Cache\UseCacheBackendTrait;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\search_api\IndexInterface;
use Drupal\search_api\LoggerTrait;
use Drupal\search_api\SearchApiException;
use Drupal\search_api_solr\SolrFieldManager;
use Drupal\search_api_solr\TypedData\SolrFieldDefinition;
use Exception;
use GuzzleHttp\ClientInterface;
use Psr\Log\LoggerInterface;

/**
 * Manages the discovery of Solr fields.
 */
class ShardFieldManager extends SolrFieldManager {

  use UseCacheBackendTrait;
  use StringTranslationTrait;
  use LoggerTrait;

  /**
   * Static cache of field definitions per Solr server.
   *
   * @var array
   */
  protected $fieldDefinitions;

  /**
   * Storage for Search API servers.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $serverStorage;

  /** @var \GuzzleHttp\ClientInterface */
  protected $httpClient;

  /**
   * Constructs a new SorFieldManager.
   *
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache_backend
   *   The cache backend.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Psr\Log\LoggerInterface $logger
   *   Logger for Search API.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   */
  public function __construct(CacheBackendInterface $cache_backend, EntityTypeManagerInterface $entityTypeManager, LoggerInterface $logger, ClientInterface $httpClient) {
    parent::__construct($cache_backend, $entityTypeManager, $logger);
    $this->httpClient = $httpClient;
  }

  /**
   * Builds the field definitions for a Solr server.
   *
   * Initially the defintions will be built from a the response of a luke query
   * handler directly from Solr. But once added to the Drupal config, the
   * definitions will be a mix of the Drupal config and not yet used fields from
   * Solr. This strategy also covers scenarios when the Solr server is
   * temporarily offline or re-indexed and prevents exceptions in Drupal's admin
   * UI.
   *
   * @param \Drupal\search_api\IndexInterface $index
   *   The index from which we are retrieving field information.
   *
   * @return \Drupal\Core\TypedData\DataDefinitionInterface[]
   *   The array of field definitions for the server, keyed by field name.
   *
   * @throws \InvalidArgumentException
   * @throws \Drupal\search_api\SearchApiException
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  protected function buildFieldDefinitions(IndexInterface $index) {
    $solr_fields = [];

    if ($this->isShardEnabled($index) === TRUE) {
      $solr_fields = $this->buildFieldDefinitionsFromSolr($index);
    }
    elseif ($this->isShardEnabled($index) === FALSE) {
      $solr_fields = parent::buildFieldDefinitionsFromSolr($index);
    }

    $config_fields = $this->buildFieldDefinitionsFromConfig($index);
    $fields = $solr_fields + $config_fields;
    /*** @var \Drupal\Core\TypedData\DataDefinitionInterface $field */
    foreach ($config_fields as $key => $field) {
      // Always use the type as already configured in Drupal previously.
      $fields[$key]->setDataType($field->getDataType());
    }
    return $fields;
  }

  protected function isShardEnabled(IndexInterface $index) {
    if ($index->getDatasources()) {
      return (bool) $index->getDatasource('solr_document')
        ->getConfiguration()['enable_shards'];
    }

    return NULL;
  }

  /**
   * Builds the field definitions for a Solr server from its Luke handler.
   *
   * @param \Drupal\search_api\IndexInterface $index
   *   The index from which we are retrieving field information.
   *
   * @return \Drupal\Core\TypedData\DataDefinitionInterface[]
   *   The array of field definitions for the server, keyed by field name.
   *
   * @throws \InvalidArgumentException
   * @throws \Drupal\search_api\SearchApiException
   */
  protected function buildFieldDefinitionsFromSolr(IndexInterface $index) {
    $fields = [];
    $shards = $this->getShardServers($index);
    $server = $index->getServerInstance();

    foreach ($shards as $shard) {
      try {
        $response = $this->httpClient->get($shard);
        $shardFields = json_decode((string) $response->getBody(), TRUE)['fields'];
        foreach ($shardFields as $name => $definition) {
          $field = new SolrFieldDefinition($definition);
          $label = Unicode::ucfirst(trim(str_replace('_', ' ', $name)));
          $field->setLabel($label);

          if (strpos($field->getDataType(), 'text') !== FALSE) {
            $field->setDataType('search_api_text');
          }
          elseif (strpos($field->getDataType(), 'date') !== FALSE) {
            $field->setDataType('timestamp');
          }
          elseif (strpos($field->getDataType(), 'int') !== FALSE) {
            $field->setDataType('integer');
          }
          elseif (strpos($field->getDataType(), 'long') !== FALSE) {
            $field->setDataType('integer');
          }
          elseif (strpos($field->getDataType(), 'float') !== FALSE) {
            $field->setDataType('float');
          }
          elseif (strpos($field->getDataType(), 'double') !== FALSE) {
            $field->setDataType('float');
          }
          elseif (strpos($field->getDataType(), 'bool') !== FALSE) {
            $field->setDataType('boolean');
          }
          else {
            $field->setDataType('string');
          }
          $fields[$name] = $field;
        }
      } catch (Exception $e) {
        $this->logger
          ->error('Could not recieve shard servers or their fields, %message', [
            '%server' => $server->id(),
            '%message' => $e->getMessage(),
          ]);
      }
    }

    return $fields;
  }

  /**
   * Extract, from solrconfig.xml solr-core file, the shards servers if exists.
   * If Solr version is major than > 6, config file may vary
   *
   * @param \Drupal\search_api\Entity\Server $server
   *
   * @return array|false|string[]
   */
  protected function getShardServers(IndexInterface $index) {
    $shards = [];

    if ($index->getServerInstance()) {
      /** @var \Drupal\search_api_solr_shards\Plugin\search_api\datasource\ShardDocument $shardDocument */
      try {
        $shardDocument = $index->getDatasource('solr_document');
        $solrconfig = empty($shardDocument->getConfiguration()['solrconfig'])
          ? 'solrconfig.xml' : $shardDocument->getConfiguration()['solrconfig'];
      }
      catch (SearchApiException $e) {
        $solrconfig = 'solrconfig.xml';
      }

      try {
        $server = $index->getServerInstance();
        $backend = $server->getBackend();
        $connector = $backend->getSolrConnector();
        $solrconfig = $connector->getFile($solrconfig)->getBody();

        $xml = simplexml_load_string($solrconfig);
        $shards = ((array) $xml->xpath('//*[@name="shards"]')[0])[0];
        if ($shards) {
          $shards = explode(',', $shards);
        }

        // Include the segment to retrieve field-schema information.
        foreach ($shards as &$shard) {
          $shard = "${shard}/admin/luke?wt=json";
        }
      } catch (SearchApiException $e) {
        $this->logger
          ->error('Could not read %file file from remote, %message', [
            '%message' => $e->getMessage(),
            '%file' => $solrconfig
          ]);
      } catch (Exception $e) {
        $this->logger
          ->error('An error occurred retrieving shards servers, %message', [
            '%message' => $e->getMessage()
          ]);
      }
    }

    return $shards;
  }

}
